# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
MLIR lowering support for union operations (getattr, setattr)
"""

from numba_cuda_mlir.type_defs.aggregate_types import UnionType, AggregateType
from numba_cuda_mlir.descriptor import MLIRTargetContext
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir.lowering_utilities import convert
from numba_cuda_mlir._mlir.dialects import llvm, arith
from numba_cuda_mlir._mlir import ir as mlir_ir
from numba_cuda_mlir.numba_cuda.core import ir as numba_ir
from numba_cuda_mlir.numba_cuda import types as numba_types


def _lower_union_construction_impl(
    context,  # Unused, kept for compatibility
    builder: "MLIRLower",
    target: numba_ir.Var,
    args: list[numba_ir.Var],
    kwargs: dict,
):
    """
    Lower union construction: MyUnion() creates a zero-initialized union value.

    A union is stored as an integer of the appropriate size (matching the largest variant).
    Access to variants is done via bitcasting.
    """
    trace(
        "Lowering union construction: target=%s, args=%s, kwargs=%s",
        target,
        args,
        kwargs,
    )

    # Get the union type being constructed
    target_type = builder.get_numba_type(target.name)
    trace("Target type: %s, type class: %s", target_type, type(target_type))

    # Get the MLIR type for the union storage
    mlir_union_type = builder.get_mlir_type(target_type)

    # Create a zero-initialized union value
    zero_union = llvm.mlir_zero(res=mlir_union_type)

    builder.store_var(target, zero_union)
    trace("Created zero-initialized union of type %s in %s", target_type, target.name)


@registry.lower(UnionType)
def lower_cunion_type_call(
    builder: "MLIRLower",
    target: numba_ir.Var,
    args: list[numba_ir.Var],
    kwargs: dict,
):
    """Lower union construction when UnionType is called directly."""
    return _lower_union_construction_impl(None, builder, target, args, kwargs)


@registry.lower_getattr_generic(UnionType)
def lower_union_type_getattr(
    context: MLIRTargetContext,
    builder: "MLIRLower",
    target: numba_ir.Var,
    value: numba_ir.Var,
    attr: str,
):
    """
    Lower getattr for union variant access.

    For unions, we bitcast the storage to the appropriate variant type.
    For struct variants, we extract bits and manually construct the struct.
    """

    union_type = builder.get_numba_type(value.name)
    trace("Lowering getattr for UnionType: %s.%s", union_type.name, attr)

    # Find the variant type
    variant_type = union_type.get_variant_type(attr)
    if variant_type is None:
        raise AttributeError(
            f"Union type '{union_type.name}' has no variant '{attr}'. "
            f"Available variants: {[v[0] for v in union_type.variants]}"
        )

    # Load the union storage
    union_value = builder.load_var(value)
    trace("Union storage type: %s", union_value.type)

    # Get the MLIR type for the variant
    variant_mlir_type = builder.get_mlir_type(variant_type)

    # Check if variant is a struct (aggregate type)
    if isinstance(variant_type, AggregateType):
        # For structs, we extract bits and construct the struct manually
        # Start with a zero-initialized struct
        variant_value = llvm.mlir_zero(res=variant_mlir_type)

        # Extract each field from the union integer
        # Separate handling for bitfield vs non-bitfield structs to avoid offset confusion

        if variant_type.is_bitfield_struct:
            # Bitfield struct: struct has a single storage field at position 0
            # Just insert the union value directly as the storage field

            # Get the storage type for this bitfield struct
            storage_type = variant_type.get_bitfield_storage_type()

            storage_mlir_type = builder.get_mlir_type(storage_type)

            # Convert union value to storage type if needed
            storage_value = convert(union_value, storage_mlir_type)

            # Insert storage value at position 0 (the only field)
            variant_value = llvm.insertvalue(
                container=variant_value,
                value=storage_value,
                position=mlir_ir.DenseI64ArrayAttr.get([0]),
            )
        else:
            # Regular (non-bitfield) struct: extract fields from union storage
            # Use field_layout (single source of truth for all field information)
            for field_name, field_info in variant_type.field_layout.items():
                # Get field info from field_layout
                field_index = field_info["field_index"]
                field_type = field_info["underlying_type"]
                field_bit_offset = field_info["bit_offset"]
                field_bits = field_info["bit_width"]

                field_mlir_type = builder.get_mlir_type(field_type)
                trace(
                    "Using field_layout for '%s': offset=%s, width=%s, index=%s",
                    field_name,
                    field_bit_offset,
                    field_bits,
                    field_index,
                )

                # Shift right to get field bits at LSB
                if field_bit_offset > 0:
                    shift_amount = arith.constant(
                        builder.get_mlir_type(union_type), field_bit_offset
                    )
                    shifted = arith.shrui(union_value, shift_amount)
                else:
                    shifted = union_value

                # Mask to extract only the field bits
                mask_value = (1 << field_bits) - 1
                mask = arith.constant(builder.get_mlir_type(union_type), mask_value)
                masked = arith.andi(shifted, mask)

                # Convert to field type if needed
                field_value = convert(masked, field_mlir_type)

                # Insert into struct using field_index from field_layout
                variant_value = llvm.insertvalue(
                    container=variant_value,
                    value=field_value,
                    position=mlir_ir.DenseI64ArrayAttr.get([field_index]),
                )

                trace(
                    "Extracted field '%s' at bit offset %s",
                    field_name,
                    field_bit_offset,
                )
    elif isinstance(variant_type, numba_types.UniTuple):
        # For UniTuple (fixed-size array), extract each element from the union storage
        trace(
            "Extracting UniTuple variant: element_type=%s, count=%s",
            variant_type.dtype,
            variant_type.count,
        )

        element_type = variant_type.dtype
        element_count = variant_type.count
        element_mlir_type = builder.get_mlir_type(element_type)

        # Get element size in bits
        if hasattr(element_type, "bitwidth"):
            element_bits = element_type.bitwidth
        else:
            raise NotImplementedError(
                f"Cannot determine bit size for UniTuple element type {element_type}"
            )

        # Extract each element
        elements = []
        union_mlir_type = builder.get_mlir_type(union_type)

        for i in range(element_count):
            bit_offset = i * element_bits

            # Shift right to get element bits at LSB
            if bit_offset > 0:
                shift_amount = arith.constant(union_mlir_type, bit_offset)
                shifted = arith.shrui(union_value, shift_amount)
            else:
                shifted = union_value

            # Mask to extract only the element bits
            mask_value = (1 << element_bits) - 1
            mask = arith.constant(union_mlir_type, mask_value)
            masked = arith.andi(shifted, mask)

            # Convert to element type if needed
            element_value = convert(masked, element_mlir_type)

            elements.append(element_value)

        # Return as Python tuple - this is how UniTuple is represented in the builder
        variant_value = tuple(elements)
        trace("Built UniTuple with %s elements", len(elements))
    else:
        variant_value = convert(union_value, variant_mlir_type)

    # Store the casted value
    builder.store_var(target, variant_value)
    trace("Extracted union variant '%s' and stored in %s", attr, target.name)




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "scf"

@_ods_cext.register_operation(_Dialect)
class ConditionOp(_ods_ir.OpView):
  r"""
  This operation accepts the continuation (i.e., inverse of exit) condition
  of the `scf.while` construct. If its first argument is true, the "after"
  region of `scf.while` is executed, with the remaining arguments forwarded
  to the entry block of the region. Otherwise, the loop terminates.
  """

  OPERATION_NAME = "scf.condition"

  _ODS_REGIONS = (0, True)

  def __init__(self, condition: _ods_ir.Value[_ods_ir.IntegerType], args: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(condition)
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def condition(self) -> _ods_ir.Value[_ods_ir.IntegerType]:
    return self.operation.operands[0]

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
    return self.operation.operands[1:1 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(ConditionOp)
class ConditionOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.condition"

  @builtins.property
  def condition(self) -> _ods_ir.Value[_ods_ir.IntegerType]:
    return self.operands[0]

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 2 + 1
    return self.operands[1:1 + _ods_variadic_group_length]

def condition(condition: _ods_ir.Value[_ods_ir.IntegerType], args: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ConditionOp:
  return ConditionOp(condition=condition, args=args, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ExecuteRegionOp(_ods_ir.OpView):
  r"""
  The `scf.execute_region` operation is used to allow multiple blocks within SCF
  and other operations which can hold only one block.  The `scf.execute_region`
  operation executes the region held exactly once and cannot have any operands.
  As such, its region has no arguments. All SSA values that dominate the op can
  be accessed inside the op. The op's region can have multiple blocks and the
  blocks can have multiple distinct terminators. Values returned from this op's
  region define the op's results.
  The optional 'no_inline' flag can be set to request the ExecuteRegionOp to be
  preserved as much as possible and not being inlined in the parent block until
  an explicit lowering step.
  
  Example:
  
  ```mlir
  scf.for %i = 0 to 128 step %c1 {
    %y = scf.execute_region -> i32 {
      %x = load %A[%i] : memref<128xi32>
      scf.yield %x : i32
    }
  }
  
  // the same as above but with no_inline attribute
  scf.for %i = 0 to 128 step %c1 {
    %y = scf.execute_region -> i32 no_inline {
      %x = load %A[%i] : memref<128xi32>
      scf.yield %x : i32
    }
  }
  
  affine.for %i = 0 to 100 {
    "foo"() : () -> ()
    %v = scf.execute_region -> i64 {
      cf.cond_br %cond, ^bb1, ^bb2
  
    ^bb1:
      %c1 = arith.constant 1 : i64
      cf.br ^bb3(%c1 : i64)
  
    ^bb2:
      %c2 = arith.constant 2 : i64
      cf.br ^bb3(%c2 : i64)
  
    ^bb3(%x : i64):
      scf.yield %x : i64
    }
    "bar"(%v) : (i64) -> ()
  }
  ```
  """

  OPERATION_NAME = "scf.execute_region"

  _ODS_REGIONS = (1, True)

  def __init__(self, result: _Sequence[_ods_ir.Type], *, no_inline: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    if bool(no_inline): attributes["no_inline"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.extend(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def no_inline(self) -> bool:
    return "no_inline" in self.operation.attributes

  @no_inline.setter
  def no_inline(self, value):
    if bool(value):
      self.operation.attributes["no_inline"] = _ods_ir.UnitAttr.get()
    elif "no_inline" in self.operation.attributes:
      del self.operation.attributes["no_inline"]

  @no_inline.deleter
  def no_inline(self):
    del self.operation.attributes["no_inline"]

  @builtins.property
  def region(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(ExecuteRegionOp)
class ExecuteRegionOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.execute_region"

  @builtins.property
  def no_inline(self) -> bool:
    return "no_inline" in self.attributes

def execute_region(result: _Sequence[_ods_ir.Type], *, no_inline: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ExecuteRegionOp]:
  op = ExecuteRegionOp(result=result, no_inline=no_inline, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class ForOp(_ods_ir.OpView):
  r"""
  The `scf.for` operation represents a loop taking 3 SSA value as operands
  that represent the lower bound, upper bound and step respectively. The
  operation defines an SSA value for its induction variable. It has one
  region capturing the loop body. The induction variable is represented as an
  argument of this region. This SSA value is a signless integer or index.
  The step is a value of same type but required to be positive, the lower and
  upper bounds can be also negative or zero. The lower and upper bounds
  specify a half-open range: the iteration is executed iff the comparison of
  induction variable value is less than the upper bound and bigger or equal
  to the lower bound.
  
  By default, the integer comparison is signed. If the `unsignedCmp` unit
  attribute is specified, the integer comparison is unsigned.
  
  The body region must contain exactly one block that terminates with
  `scf.yield`. Calling ForOp::build will create such a region and insert
  the terminator implicitly if none is defined, so will the parsing even in
  cases when it is absent from the custom format. For example:
  
  ```mlir
  // Index case.
  scf.for %iv = %lb to %ub step %step {
    ... // body
  }
  ...
  // Unsigned integer case.
  scf.for unsigned %iv_32 = %lb_32 to %ub_32 step %step_32 : i32 {
    ... // body
  }
  ```
  
  `scf.for` can also operate on loop-carried variables and returns the final
  values after loop termination. The initial values of the variables are
  passed as additional SSA operands to the `scf.for` following the 3 loop
  control SSA values mentioned above (lower bound, upper bound and step). The
  operation region has an argument for the induction variable, followed by
  one argument for each loop-carried variable, representing the value of the
  variable at the current iteration.
  
  The region must terminate with a `scf.yield` that passes the current
  values of all loop-carried variables to the next iteration, or to the
  `scf.for` result, if at the last iteration. The static type of a
  loop-carried variable may not change with iterations; its runtime type is
  allowed to change. Note, that when the loop-carried variables are present,
  calling ForOp::build will not insert the terminator implicitly. The caller
  must insert `scf.yield` in that case.
  
  `scf.for` results hold the final values after the last iteration.
  For example, to sum-reduce a memref:
  
  ```mlir
  func.func @reduce(%buffer: memref<1024xf32>, %lb: index,
                    %ub: index, %step: index) -> (f32) {
    // Initial sum set to 0.
    %sum_0 = arith.constant 0.0 : f32
    // iter_args binds initial values to the loop's region arguments.
    %sum = scf.for %iv = %lb to %ub step %step
        iter_args(%sum_iter = %sum_0) -> (f32) {
      %t = load %buffer[%iv] : memref<1024xf32>
      %sum_next = arith.addf %sum_iter, %t : f32
      // Yield current iteration sum to next iteration %sum_iter or to %sum
      // if final iteration.
      scf.yield %sum_next : f32
    }
    return %sum : f32
  }
  ```
  
  If the `scf.for` defines any values, a yield must be explicitly present.
  The number and types of the `scf.for` results must match the initial
  values in the `iter_args` binding and the yield operands.
  
  Another example with a nested `scf.if` (see `scf.if` for details) to
  perform conditional reduction:
  
  ```mlir
  func.func @conditional_reduce(%buffer: memref<1024xf32>, %lb: index,
                                %ub: index, %step: index) -> (f32) {
    %sum_0 = arith.constant 0.0 : f32
    %c0 = arith.constant 0.0 : f32
    %sum = scf.for %iv = %lb to %ub step %step
        iter_args(%sum_iter = %sum_0) -> (f32) {
      %t = load %buffer[%iv] : memref<1024xf32>
      %cond = arith.cmpf "ugt", %t, %c0 : f32
      %sum_next = scf.if %cond -> (f32) {
        %new_sum = arith.addf %sum_iter, %t : f32
        scf.yield %new_sum : f32
      } else {
        scf.yield %sum_iter : f32
      }
      scf.yield %sum_next : f32
    }
    return %sum : f32
  }
  ```
  """

  OPERATION_NAME = "scf.for"

  _ODS_REGIONS = (1, True)

  def __init__(self, results_: _Sequence[_ods_ir.Type], lowerBound: _ods_ir.Value, upperBound: _ods_ir.Value, step: _ods_ir.Value, initArgs: _Sequence[_ods_ir.Value], *, unsignedCmp: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lowerBound)
    operands.append(upperBound)
    operands.append(step)
    operands.extend(_get_op_results_or_values(initArgs))
    _ods_context = _ods_get_default_loc_context(loc)
    if bool(unsignedCmp): attributes["unsignedCmp"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.extend(results_)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lowerBound(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def upperBound(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def step(self) -> _ods_ir.Value:
    return self.operation.operands[2]

  @builtins.property
  def initArgs(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 4 + 1
    return self.operation.operands[3:3 + _ods_variadic_group_length]

  @builtins.property
  def unsignedCmp(self) -> bool:
    return "unsignedCmp" in self.operation.attributes

  @unsignedCmp.setter
  def unsignedCmp(self, value):
    if bool(value):
      self.operation.attributes["unsignedCmp"] = _ods_ir.UnitAttr.get()
    elif "unsignedCmp" in self.operation.attributes:
      del self.operation.attributes["unsignedCmp"]

  @unsignedCmp.deleter
  def unsignedCmp(self):
    del self.operation.attributes["unsignedCmp"]

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def region(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(ForOp)
class ForOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.for"

  @builtins.property
  def lowerBound(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def upperBound(self) -> _ods_ir.Value:
    return self.operands[1]

  @builtins.property
  def step(self) -> _ods_ir.Value:
    return self.operands[2]

  @builtins.property
  def initArgs(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 4 + 1
    return self.operands[3:3 + _ods_variadic_group_length]

  @builtins.property
  def unsignedCmp(self) -> bool:
    return "unsignedCmp" in self.attributes

def for_(results_: _Sequence[_ods_ir.Type], lower_bound: _ods_ir.Value, upper_bound: _ods_ir.Value, step: _ods_ir.Value, init_args: _Sequence[_ods_ir.Value], *, unsigned_cmp: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ForOp]:
  op = ForOp(results_=results_, lowerBound=lower_bound, upperBound=upper_bound, step=step, initArgs=init_args, unsignedCmp=unsigned_cmp, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class ForallOp(_ods_ir.OpView):
  r"""
  `scf.forall` is a target-independent multi-dimensional parallel
  region application operation. It has exactly one block that represents the
  parallel body and it takes index operands that specify lower bounds, upper
  bounds and steps.
  
  The op also takes a variadic number of tensor operands (`shared_outs`).
  The future buffers corresponding to these tensors are shared among all
  threads. Shared tensors should be accessed via their corresponding block
  arguments. If multiple threads write to a shared buffer in a racy
  fashion, these writes will execute in some unspecified order. Tensors that
  are not shared can be used inside the body (i.e., the op is not isolated
  from above); however, if a use of such a tensor bufferizes to a memory
  write, the tensor is privatized, i.e., a thread-local copy of the tensor is
  used. This ensures that memory side effects of a thread are not visible to
  other threads (or in the parent body), apart from explicitly shared tensors.
  
  The name "thread" conveys the fact that the parallel execution is mapped
  (i.e. distributed) to a set of virtual threads of execution, one function
  application per thread. Further lowerings are responsible for specifying
  how this is materialized on concrete hardware resources.
  
  An optional `mapping` is an attribute array that specifies processing units
  with their dimension, how it remaps 1-1 to a set of concrete processing
  element resources (e.g. a CUDA grid dimension or a level of concrete nested
  async parallelism). It is expressed via any attribute that implements the
  device mapping interface. It is the reponsibility of the lowering mechanism
  to interpret the `mapping` attributes in the context of the concrete target
  the op is lowered to, or to ignore it when the specification is ill-formed
  or unsupported for a particular target.
  
  The only allowed terminator is `scf.forall.in_parallel`.
  `scf.forall` returns one value per `shared_out` operand. The
  actions of the `scf.forall.in_parallel` terminators specify how to combine the
  partial results of all parallel invocations into a full value, in some
  unspecified order. The "destination" of each such op must be a `shared_out`
  block argument of the `scf.forall` op.
  
  The actions involved in constructing the return values are further described
  by `tensor.parallel_insert_slice`.
  
  `scf.forall` acts as an implicit synchronization point.
  
  When the parallel function body has side effects, their order is unspecified
  across threads.
  
  `scf.forall` can be printed in two different ways depending on
  whether the loop is normalized or not. The loop is 'normalized' when all
  lower bounds are equal to zero and steps are equal to one. In that case,
  `lowerBound` and `step` operands will be omitted during printing.
  
  Normalized loop example:
  
  ```mlir
  //
  // Sequential context.
  //
  %matmul_and_pointwise:2 = scf.forall (%thread_id_1, %thread_id_2) in
      (%num_threads_1, %numthread_id_2) shared_outs(%o1 = %C, %o2 = %pointwise)
    -> (tensor<?x?xT>, tensor<?xT>) {
    //
    // Parallel context, each thread with id = (%thread_id_1, %thread_id_2)
    // runs its version of the code.
    //
    %sA = tensor.extract_slice %A[f((%thread_id_1, %thread_id_2))]:
      tensor<?x?xT> to tensor<?x?xT>
    %sB = tensor.extract_slice %B[g((%thread_id_1, %thread_id_2))]:
      tensor<?x?xT> to tensor<?x?xT>
    %sC = tensor.extract_slice %o1[h((%thread_id_1, %thread_id_2))]:
      tensor<?x?xT> to tensor<?x?xT>
    %sD = linalg.matmul
      ins(%sA, %sB : tensor<?x?xT>, tensor<?x?xT>)
      outs(%sC : tensor<?x?xT>)
  
    %spointwise = subtensor %o2[i((%thread_id_1, %thread_id_2))]:
      tensor<?xT> to tensor<?xT>
    %sE = linalg.add ins(%spointwise : tensor<?xT>) outs(%sD : tensor<?xT>)
  
    scf.forall.in_parallel {
      tensor.parallel_insert_slice %sD into %o1[h((%thread_id_1, %thread_id_2))]:
        tensor<?x?xT> into tensor<?x?xT>
  
      tensor.parallel_insert_slice %spointwise into %o2[i((%thread_id_1, %thread_id_2))]:
        tensor<?xT> into tensor<?xT>
    }
  }
  // Implicit synchronization point.
  // Sequential context.
  //
  ```
  
  Loop with loop bounds example:
  
  ```mlir
  //
  // Sequential context.
  //
  %pointwise = scf.forall (%i, %j) = (0, 0) to (%dim1, %dim2)
    step (%tileSize1, %tileSize2) shared_outs(%o1 = %out)
    -> (tensor<?x?xT>, tensor<?xT>) {
    //
    // Parallel context.
    //
    %sA = tensor.extract_slice %A[%i, %j][%tileSize1, %tileSize2][1, 1]
      : tensor<?x?xT> to tensor<?x?xT>
    %sB = tensor.extract_slice %B[%i, %j][%tileSize1, %tileSize2][1, 1]
      : tensor<?x?xT> to tensor<?x?xT>
    %sC = tensor.extract_slice %o[%i, %j][%tileSize1, %tileSize2][1, 1]
      : tensor<?x?xT> to tensor<?x?xT>
  
    %add = linalg.map {"arith.addf"}
      ins(%sA, %sB : tensor<?x?xT>, tensor<?x?xT>)
      outs(%sC : tensor<?x?xT>)
  
    scf.forall.in_parallel {
      tensor.parallel_insert_slice %add into
        %o[%i, %j][%tileSize1, %tileSize2][1, 1]
        : tensor<?x?xT> into tensor<?x?xT>
    }
  }
  // Implicit synchronization point.
  // Sequential context.
  //
  ```
  
  Example with mapping attribute:
  
  ```mlir
  //
  // Sequential context. Here `mapping` is expressed as GPU thread mapping
  // attributes
  //
  %matmul_and_pointwise:2 = scf.forall (%thread_id_1, %thread_id_2) in
      (%num_threads_1, %numthread_id_2) shared_outs(...)
    -> (tensor<?x?xT>, tensor<?xT>) {
    //
    // Parallel context, each thread with id = **(%thread_id_2, %thread_id_1)**
    // runs its version of the code.
    //
     scf.forall.in_parallel {
       ...
    }
  } { mapping = [#gpu.thread<y>, #gpu.thread<x>] }
  // Implicit synchronization point.
  // Sequential context.
  //
  ```
  
  Example with privatized tensors:
  
  ```mlir
  %t0 = ...
  %t1 = ...
  %r = scf.forall ... shared_outs(%o = t0) -> tensor<?xf32> {
    // %t0 and %t1 are privatized. %t0 is definitely copied for each thread
    // because the scf.forall op's %t0 use bufferizes to a memory
    // write. In the absence of other conflicts, %t1 is copied only if there
    // are uses of %t1 in the body that bufferize to a memory read and to a
    // memory write.
    "some_use"(%t0)
    "some_use"(%t1)
  }
  ```
  """

  OPERATION_NAME = "scf.forall"

  _ODS_OPERAND_SEGMENTS = [-1,-1,-1,-1,]

  _ODS_REGIONS = (1, True)

  def __init__(self, results_: _Sequence[_ods_ir.Type], dynamicLowerBound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], dynamicUpperBound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], dynamicStep: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], staticLowerBound: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], staticUpperBound: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], staticStep: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], outputs: _Sequence[_ods_ir.Value[_ods_ir.RankedTensorType]], *, mapping: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(_get_op_results_or_values(dynamicLowerBound))
    operands.append(_get_op_results_or_values(dynamicUpperBound))
    operands.append(_get_op_results_or_values(dynamicStep))
    operands.append(_get_op_results_or_values(outputs))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["staticLowerBound"] = (staticLowerBound if (
    isinstance(staticLowerBound, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('DenseI64ArrayAttr')) else
      _ods_ir.AttrBuilder.get('DenseI64ArrayAttr')(staticLowerBound, context=_ods_context))
    attributes["staticUpperBound"] = (staticUpperBound if (
    isinstance(staticUpperBound, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('DenseI64ArrayAttr')) else
      _ods_ir.AttrBuilder.get('DenseI64ArrayAttr')(staticUpperBound, context=_ods_context))
    attributes["staticStep"] = (staticStep if (
    isinstance(staticStep, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('DenseI64ArrayAttr')) else
      _ods_ir.AttrBuilder.get('DenseI64ArrayAttr')(staticStep, context=_ods_context))
    if mapping is not None: attributes["mapping"] = (mapping if (
        isinstance(mapping, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DeviceMappingArrayAttr')) else
          _ods_ir.AttrBuilder.get('DeviceMappingArrayAttr')(mapping, context=_ods_context))
    results = []
    results.extend(results_)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def dynamicLowerBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def dynamicUpperBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def dynamicStep(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 2)
    return operand_range

  @builtins.property
  def outputs(self) -> _ods_ir.OpOperandList[_ods_ir.RankedTensorType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 3)
    return operand_range

  @builtins.property
  def staticLowerBound(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.operation.attributes["staticLowerBound"]

  @staticLowerBound.setter
  def staticLowerBound(self, value: _ods_ir.DenseI64ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["staticLowerBound"] = value

  @builtins.property
  def staticUpperBound(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.operation.attributes["staticUpperBound"]

  @staticUpperBound.setter
  def staticUpperBound(self, value: _ods_ir.DenseI64ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["staticUpperBound"] = value

  @builtins.property
  def staticStep(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.operation.attributes["staticStep"]

  @staticStep.setter
  def staticStep(self, value: _ods_ir.DenseI64ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["staticStep"] = value

  @builtins.property
  def mapping(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "mapping" not in self.operation.attributes:
      return None
    return self.operation.attributes["mapping"]

  @mapping.setter
  def mapping(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["mapping"] = value
    elif "mapping" in self.operation.attributes:
      del self.operation.attributes["mapping"]

  @mapping.deleter
  def mapping(self):
    del self.operation.attributes["mapping"]

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def region(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(ForallOp)
class ForallOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.forall"

  @builtins.property
  def dynamicLowerBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def dynamicUpperBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def dynamicStep(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 2)
    return operand_range

  @builtins.property
  def outputs(self) -> _ods_ir.OpOperandList[_ods_ir.RankedTensorType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 3)
    return operand_range

  @builtins.property
  def staticLowerBound(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.attributes["staticLowerBound"]

  @builtins.property
  def staticUpperBound(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.attributes["staticUpperBound"]

  @builtins.property
  def staticStep(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.attributes["staticStep"]

  @builtins.property
  def mapping(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "mapping" not in self.attributes:
      return None
    return self.attributes["mapping"]

def forall(results_: _Sequence[_ods_ir.Type], dynamic_lower_bound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], dynamic_upper_bound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], dynamic_step: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], static_lower_bound: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], static_upper_bound: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], static_step: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], outputs: _Sequence[_ods_ir.Value[_ods_ir.RankedTensorType]], *, mapping: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ForallOp]:
  op = ForallOp(results_=results_, dynamicLowerBound=dynamic_lower_bound, dynamicUpperBound=dynamic_upper_bound, dynamicStep=dynamic_step, staticLowerBound=static_lower_bound, staticUpperBound=static_upper_bound, staticStep=static_step, outputs=outputs, mapping=mapping, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class IfOp(_ods_ir.OpView):
  r"""
  The `scf.if` operation represents an if-then-else construct for
  conditionally executing two regions of code. The operand to an if operation
  is a boolean value. For example:
  
  ```mlir
  scf.if %b  {
    ...
  } else {
    ...
  }
  ```
  
  `scf.if` may also produce results. Which values are returned depends on
  which execution path is taken.
  
  Example:
  
  ```mlir
  %x, %y = scf.if %b -> (f32, f32) {
    %x_true = ...
    %y_true = ...
    scf.yield %x_true, %y_true : f32, f32
  } else {
    %x_false = ...
    %y_false = ...
    scf.yield %x_false, %y_false : f32, f32
  }
  ```
  
  The "then" region has exactly 1 block. The "else" region may have 0 or 1
  block. In case the `scf.if` produces results, the "else" region must also
  have exactly 1 block.
  
  The blocks are always terminated with `scf.yield`. If `scf.if` defines no
  values, the `scf.yield` can be left out, and will be inserted implicitly.
  Otherwise, it must be explicit.
  
  Example:
  
  ```mlir
  scf.if %b  {
    ...
  }
  ```
  
  The types of the yielded values must match the result types of the
  `scf.if`.
  """

  OPERATION_NAME = "scf.if"

  _ODS_REGIONS = (2, True)

  def __init__(self, results_: _Sequence[_ods_ir.Type], condition: _ods_ir.Value[_ods_ir.IntegerType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(condition)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.extend(results_)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def condition(self) -> _ods_ir.Value[_ods_ir.IntegerType]:
    return self.operation.operands[0]

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def thenRegion(self) -> _ods_ir.Region:
    return self.regions[0]

  @builtins.property
  def elseRegion(self) -> _ods_ir.Region:
    return self.regions[1]

@_ods_cext.register_op_adaptor(IfOp)
class IfOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.if"

  @builtins.property
  def condition(self) -> _ods_ir.Value[_ods_ir.IntegerType]:
    return self.operands[0]

def if_(results_: _Sequence[_ods_ir.Type], condition: _ods_ir.Value[_ods_ir.IntegerType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, IfOp]:
  op = IfOp(results_=results_, condition=condition, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class InParallelOp(_ods_ir.OpView):
  r"""
  The `scf.forall.in_parallel` is a designated terminator for
  the `scf.forall` operation.
  
  It has a single region with a single block that contains a flat list of ops.
  Each such op participates in the aggregate formation of a single result of
  the enclosing `scf.forall`.
  The result number corresponds to the position of the op in the terminator.
  """

  OPERATION_NAME = "scf.forall.in_parallel"

  _ODS_REGIONS = (1, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def region(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(InParallelOp)
class InParallelOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.forall.in_parallel"

def forall_in_parallel(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> InParallelOp:
  return InParallelOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class IndexSwitchOp(_ods_ir.OpView):
  r"""
  The `scf.index_switch` is a control-flow operation that branches to one of
  the given regions based on the values of the argument and the cases. The
  argument is always of type `index`.
  
  The operation always has a "default" region and any number of case regions
  denoted by integer constants. Control-flow transfers to the case region
  whose constant value equals the value of the argument. If the argument does
  not equal any of the case values, control-flow transfer to the "default"
  region.
  
  Example:
  
  ```mlir
  %0 = scf.index_switch %arg0 : index -> i32
  case 2 {
    %1 = arith.constant 10 : i32
    scf.yield %1 : i32
  }
  case 5 {
    %2 = arith.constant 20 : i32
    scf.yield %2 : i32
  }
  default {
    %3 = arith.constant 30 : i32
    scf.yield %3 : i32
  }
  ```
  """

  OPERATION_NAME = "scf.index_switch"

  _ODS_REGIONS = (1, False)

  def __init__(self, results_: _Sequence[_ods_ir.Type], arg: _ods_ir.Value[_ods_ir.IndexType], cases: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], num_caseRegions: int, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(arg)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["cases"] = (cases if (
    isinstance(cases, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('DenseI64ArrayAttr')) else
      _ods_ir.AttrBuilder.get('DenseI64ArrayAttr')(cases, context=_ods_context))
    results = []
    results.extend(results_)
    _ods_successors = None
    regions = 1 + num_caseRegions
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def arg(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def cases(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.operation.attributes["cases"]

  @cases.setter
  def cases(self, value: _ods_ir.DenseI64ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["cases"] = value

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def defaultRegion(self) -> _ods_ir.Region:
    return self.regions[0]

  @builtins.property
  def caseRegions(self) -> _ods_ir.RegionSequence:
    return self.regions[1:]

@_ods_cext.register_op_adaptor(IndexSwitchOp)
class IndexSwitchOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.index_switch"

  @builtins.property
  def arg(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def cases(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.attributes["cases"]

def index_switch(results_: _Sequence[_ods_ir.Type], arg: _ods_ir.Value[_ods_ir.IndexType], cases: _Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr], num_case_regions: int, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, IndexSwitchOp]:
  op = IndexSwitchOp(results_=results_, arg=arg, cases=cases, num_caseRegions=num_case_regions, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class ParallelOp(_ods_ir.OpView):
  r"""
  The `scf.parallel` operation represents a loop nest taking 4 groups of SSA
  values as operands that represent the lower bounds, upper bounds, steps and
  initial values, respectively. The operation defines a variadic number of
  SSA values for its induction variables. It has one region capturing the
  loop body. The induction variables are represented as an argument of this
  region. These SSA values always have type index, which is the size of the
  machine word. The steps are values of type index, required to be positive.
  The lower and upper bounds specify a half-open range: the range includes
  the lower bound but does not include the upper bound. The initial values
  have the same types as results of `scf.parallel`. If there are no results,
  the keyword `init` can be omitted.
  
  Semantically we require that the iteration space can be iterated in any
  order, and the loop body can be executed in parallel. If there are data
  races, the behavior is undefined.
  
  The parallel loop operation supports reduction of values produced by
  individual iterations into a single result. This is modeled using the
  `scf.reduce` terminator operation (see `scf.reduce` for details). The i-th
  result of an `scf.parallel` operation is associated with the i-th initial
  value operand, the i-th operand of the `scf.reduce` operation (the value to
  be reduced) and the i-th region of the `scf.reduce` operation (the reduction
  function). Consequently, we require that the number of results of an
  `scf.parallel` op matches the number of initial values and the the number of
  reductions in the `scf.reduce` terminator.
  
  The body region must contain exactly one block that terminates with a
  `scf.reduce` operation. If an `scf.parallel` op has no reductions, the
  terminator has no operands and no regions. The `scf.parallel` parser will
  automatically insert the terminator for ops that have no reductions if it is
  absent.
  
  Example:
  
  ```mlir
  %init = arith.constant 0.0 : f32
  %r:2 = scf.parallel (%iv) = (%lb) to (%ub) step (%step) init (%init, %init)
      -> f32, f32 {
    %elem_to_reduce1 = load %buffer1[%iv] : memref<100xf32>
    %elem_to_reduce2 = load %buffer2[%iv] : memref<100xf32>
    scf.reduce(%elem_to_reduce1, %elem_to_reduce2 : f32, f32) {
      ^bb0(%lhs : f32, %rhs: f32):
        %res = arith.addf %lhs, %rhs : f32
        scf.reduce.return %res : f32
    }, {
      ^bb0(%lhs : f32, %rhs: f32):
        %res = arith.mulf %lhs, %rhs : f32
        scf.reduce.return %res : f32
    }
  }
  ```
  """

  OPERATION_NAME = "scf.parallel"

  _ODS_OPERAND_SEGMENTS = [-1,-1,-1,-1,]

  _ODS_REGIONS = (1, True)

  def __init__(self, results_: _Sequence[_ods_ir.Type], lowerBound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], upperBound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], step: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], initVals: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(_get_op_results_or_values(lowerBound))
    operands.append(_get_op_results_or_values(upperBound))
    operands.append(_get_op_results_or_values(step))
    operands.append(_get_op_results_or_values(initVals))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.extend(results_)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lowerBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def upperBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def step(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 2)
    return operand_range

  @builtins.property
  def initVals(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 3)
    return operand_range

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def region(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(ParallelOp)
class ParallelOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.parallel"

  @builtins.property
  def lowerBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def upperBound(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def step(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 2)
    return operand_range

  @builtins.property
  def initVals(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 3)
    return operand_range

def parallel(results_: _Sequence[_ods_ir.Type], lower_bound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], upper_bound: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], step: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], init_vals: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ParallelOp]:
  op = ParallelOp(results_=results_, lowerBound=lower_bound, upperBound=upper_bound, step=step, initVals=init_vals, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class ReduceOp(_ods_ir.OpView):
  r"""
  The `scf.reduce` operation is the terminator for `scf.parallel` operations. It can model
  an arbitrary number of reductions. It has one region per reduction. Each
  region has one block with two arguments which have the same type as the
  corresponding operand of `scf.reduce`. The operands of the op are the values
  that should be reduce; one value per reduction.
  
  The i-th reduction (i.e., the i-th region and the i-th operand) corresponds
  the i-th initial value and the i-th result of the enclosing `scf.parallel`
  op.
  
  The `scf.reduce` operation contains regions whose entry blocks expect two
  arguments of the same type as the corresponding operand. As the iteration
  order of the enclosing parallel loop and hence reduction order is
  unspecified, the results of the reductions may be non-deterministic unless
  the reductions are associative and commutative.
  
  The result of a reduction region (`scf.reduce.return` operand) must have the
  same type as the corresponding `scf.reduce` operand and the corresponding
  `scf.parallel` initial value.
  
  Example:
  
  ```mlir
  %operand = arith.constant 1.0 : f32
  scf.reduce(%operand : f32) {
    ^bb0(%lhs : f32, %rhs: f32):
      %res = arith.addf %lhs, %rhs : f32
      scf.reduce.return %res : f32
  }
  ```
  """

  OPERATION_NAME = "scf.reduce"

  _ODS_REGIONS = (0, False)

  def __init__(self, operands_: _Sequence[_ods_ir.Value], num_reductions: int, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(operands_))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    regions = 0 + num_reductions
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def reductions(self) -> _ods_ir.RegionSequence:
    return self.regions[0:]

@_ods_cext.register_op_adaptor(ReduceOp)
class ReduceOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.reduce"

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def reduce(operands_: _Sequence[_ods_ir.Value], num_reductions: int, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ReduceOp:
  return ReduceOp(operands_=operands_, num_reductions=num_reductions, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ReduceReturnOp(_ods_ir.OpView):
  r"""
  The `scf.reduce.return` operation is a special terminator operation for the block inside
  `scf.reduce` regions. It terminates the region. It should have the same
  operand type as the corresponding operand of the enclosing `scf.reduce` op.
  
  Example:
  
  ```mlir
  scf.reduce.return %res : f32
  ```
  """

  OPERATION_NAME = "scf.reduce.return"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(result)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def result(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(ReduceReturnOp)
class ReduceReturnOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.reduce.return"

  @builtins.property
  def result(self) -> _ods_ir.Value:
    return self.operands[0]

def reduce_return(result: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ReduceReturnOp:
  return ReduceReturnOp(result=result, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class WhileOp(_ods_ir.OpView):
  r"""
  This operation represents a generic "while"/"do-while" loop that keeps
  iterating as long as a condition is satisfied. There is no restriction on
  the complexity of the condition. It consists of two regions (with single
  block each): "before" region and "after" region. The names of regions
  indicates whether they execute before or after the condition check.
  Therefore, if the main loop payload is located in the "before" region, the
  operation is a "do-while" loop. Otherwise, it is a "while" loop.
  
  The "before" region terminates with a special operation, `scf.condition`,
  that accepts as its first operand an `i1` value indicating whether to
  proceed to the "after" region (value is `true`) or not. The two regions
  communicate by means of region arguments. Initially, the "before" region
  accepts as arguments the operands of the `scf.while` operation and uses them
  to evaluate the condition. It forwards the trailing, non-condition operands
  of the `scf.condition` terminator either to the "after" region if the
  control flow is transferred there or to results of the `scf.while` operation
  otherwise. The "after" region takes as arguments the values produced by the
  "before" region and uses `scf.yield` to supply new arguments for the
  "before" region, into which it transfers the control flow unconditionally.
  
  A simple "while" loop can be represented as follows.
  
  ```mlir
  %res = scf.while (%arg1 = %init1) : (f32) -> f32 {
    // "Before" region.
    // In a "while" loop, this region computes the condition.
    %condition = call @evaluate_condition(%arg1) : (f32) -> i1
  
    // Forward the argument (as result or "after" region argument).
    scf.condition(%condition) %arg1 : f32
  
  } do {
  ^bb0(%arg2: f32):
    // "After" region.
    // In a "while" loop, this region is the loop body.
    %next = call @payload(%arg2) : (f32) -> f32
  
    // Forward the new value to the "before" region.
    // The operand types must match the types of the `scf.while` operands.
    scf.yield %next : f32
  }
  ```
  
  A simple "do-while" loop can be represented by reducing the "after" block
  to a simple forwarder.
  
  ```mlir
  %res = scf.while (%arg1 = %init1) : (f32) -> f32 {
    // "Before" region.
    // In a "do-while" loop, this region contains the loop body.
    %next = call @payload(%arg1) : (f32) -> f32
  
    // And also evaluates the condition.
    %condition = call @evaluate_condition(%arg1) : (f32) -> i1
  
    // Loop through the "after" region.
    scf.condition(%condition) %next : f32
  
  } do {
  ^bb0(%arg2: f32):
    // "After" region.
    // Forwards the values back to "before" region unmodified.
    scf.yield %arg2 : f32
  }
  ```
  
  Note that the types of region arguments need not to match with each other.
  The op expects the operand types to match with argument types of the
  "before" region; the result types to match with the trailing operand types
  of the terminator of the "before" region, and with the argument types of the
  "after" region. The following scheme can be used to share the results of
  some operations executed in the "before" region with the "after" region,
  avoiding the need to recompute them.
  
  ```mlir
  %res = scf.while (%arg1 = %init1) : (f32) -> i64 {
    // One can perform some computations, e.g., necessary to evaluate the
    // condition, in the "before" region and forward their results to the
    // "after" region.
    %shared = call @shared_compute(%arg1) : (f32) -> i64
  
    // Evaluate the condition.
    %condition = call @evaluate_condition(%arg1, %shared) : (f32, i64) -> i1
  
    // Forward the result of the shared computation to the "after" region.
    // The types must match the arguments of the "after" region as well as
    // those of the `scf.while` results.
    scf.condition(%condition) %shared : i64
  
  } do {
  ^bb0(%arg2: i64) {
    // Use the partial result to compute the rest of the payload in the
    // "after" region.
    %res = call @payload(%arg2) : (i64) -> f32
  
    // Forward the new value to the "before" region.
    // The operand types must match the types of the `scf.while` operands.
    scf.yield %res : f32
  }
  ```
  
  The custom syntax for this operation is as follows.
  
  ```
  op ::= `scf.while` assignments `:` function-type region `do` region
         `attributes` attribute-dict
  initializer ::= /* empty */ | `(` assignment-list `)`
  assignment-list ::= assignment | assignment `,` assignment-list
  assignment ::= ssa-value `=` ssa-value
  ```
  """

  OPERATION_NAME = "scf.while"

  _ODS_REGIONS = (2, True)

  def __init__(self, results_: _Sequence[_ods_ir.Type], inits: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(inits))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.extend(results_)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def inits(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def before(self) -> _ods_ir.Region:
    return self.regions[0]

  @builtins.property
  def after(self) -> _ods_ir.Region:
    return self.regions[1]

@_ods_cext.register_op_adaptor(WhileOp)
class WhileOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.while"

  @builtins.property
  def inits(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def while_(results_: _Sequence[_ods_ir.Type], inits: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, WhileOp]:
  op = WhileOp(results_=results_, inits=inits, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class YieldOp(_ods_ir.OpView):
  r"""
  The `scf.yield` operation yields an SSA value from the SCF dialect op region and
  terminates the regions. The semantics of how the values are yielded is
  defined by the parent operation.
  If `scf.yield` has any operands, the operands must match the parent
  operation's results.
  If the parent operation defines no values, then the `scf.yield` may be
  left out in the custom syntax and the builders will insert one implicitly.
  Otherwise, it has to be present in the syntax to indicate which values are
  yielded.
  """

  OPERATION_NAME = "scf.yield"

  _ODS_REGIONS = (0, True)

  def __init__(self, results_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(results_))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def results_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(YieldOp)
class YieldOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "scf.yield"

  @builtins.property
  def results_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def yield_(results_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> YieldOp:
  return YieldOp(results_=results_, loc=loc, ip=ip)




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "bufferization"

@_ods_cext.register_operation(_Dialect)
class AllocTensorOp(_ods_ir.OpView):
  r"""
  `bufferization.alloc_tensor` materializes an uninitialized tensor with a
  given shape (dynamic or static). It always bufferizes to a new buffer
  allocation of the given shape. The optional `copy` operand specifies the
  contents of the tensors. If no `copy` operand is specified, reading from the
  result of an `alloc_tensor` op yields an undefined value.
  
  If `copy` is specified, no dynamic sizes should be passed, since they are
  the same as the dynamic sizes of the `copy` operand.
  
  `alloc_tensor` is a helper op for bufferization. The operation is provided
  as an anchor that marks the beginning of a new tensor SSA use-def chain. It
  can be used to control in-place bufferization decisions during One-Shot
  Bufferize: The bufferized result of a `bufferization.alloc_tensor` does not
  alias with any other buffer, so it can be used to resolve read-after-write
  conflicts that would have been introduced by the in-place bufferization of
  another op.
  
  The optional `memory_space` attribute specifies the memory space when
  bufferizing this op. The memory space is inferred from `copy` if specified.
  If neither `copy` nor `memory_space` is specified, the default memory space
  is used during bufferization.
  
  The optional `size_hint` operand specifies the number of non-zero elements
  for sparse tensors. The value of `size_hint` should be not less than 1 and
  not larger than the linear size of the corresponding dense tensor type. If
  this requirement is not met, the behavior of the operator is undefined.
  
  Both dense and sparse tensor types are supported. The result of a
  `bufferization.alloc_tensor` is a tensor value that can be used like any
  other tensor value. In practice, it is often used as the "out" operand of
  another op. Sparse tensor allocations should always be used in a local
  construction operation and never escape the function boundary directly.
  
  Example:
  
  ```mlir
  %c = bufferization.alloc_tensor(%d1, %d2) : tensor<?x?xf32, #SparseMatrix>
  %0 = linalg.matmul
    ins(%a, %b: tensor<?x?xf32, #SparseMatrix>, tensor<?x?xf32, #SparseMatrix>)
    outs(%c: tensor<?x?xf32, #SparseMatrix>) -> tensor<?x?xf32, #SparseMatrix>
  return %0 : tensor<?x?xf32, #SparseMatrix>
  ```
  
  ```mlir
  %c = bufferization.alloc_tensor(%d1, %d2) size_hint = %noe
    : tensor<?x?xf32, #SparseMatrix>
  ```
  
  Note: An `alloc_tensor` with a `copy` should also be expressed as an
  `alloc_tensor` without `copy`, followed by a `copy_tensor`.
  """

  OPERATION_NAME = "bufferization.alloc_tensor"

  _ODS_OPERAND_SEGMENTS = [-1,0,0,]

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, dynamic_sizes: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], *, copy: _Optional[_ods_ir.Value] = None, size_hint: _Optional[_ods_ir.Value[_ods_ir.IndexType]] = None, memory_space: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(_get_op_results_or_values(dynamic_sizes))
    operands.append(copy)
    operands.append(size_hint)
    _ods_context = _ods_get_default_loc_context(loc)
    if memory_space is not None: attributes["memory_space"] = (memory_space if (
        isinstance(memory_space, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('AnyAttr')) else
          _ods_ir.AttrBuilder.get('AnyAttr')(memory_space, context=_ods_context))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def dynamic_sizes(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def copy(self) -> _Optional[_ods_ir.Value]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def size_hint(self) -> _Optional[_ods_ir.Value[_ods_ir.IndexType]]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 2)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def memory_space(self) -> _Optional[_ods_ir.Attribute]:
    if "memory_space" not in self.operation.attributes:
      return None
    return self.operation.attributes["memory_space"]

  @memory_space.setter
  def memory_space(self, value: _Optional[_ods_ir.Attribute]):
    if value is not None:
      self.operation.attributes["memory_space"] = value
    elif "memory_space" in self.operation.attributes:
      del self.operation.attributes["memory_space"]

  @memory_space.deleter
  def memory_space(self):
    del self.operation.attributes["memory_space"]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(AllocTensorOp)
class AllocTensorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "bufferization.alloc_tensor"

  @builtins.property
  def dynamic_sizes(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def copy(self) -> _Optional[_ods_ir.Value]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def size_hint(self) -> _Optional[_ods_ir.Value[_ods_ir.IndexType]]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 2)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def memory_space(self) -> _Optional[_ods_ir.Attribute]:
    if "memory_space" not in self.attributes:
      return None
    return self.attributes["memory_space"]

def alloc_tensor(result: _ods_ir.Type, dynamic_sizes: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], *, copy: _Optional[_ods_ir.Value] = None, size_hint: _Optional[_ods_ir.Value[_ods_ir.IndexType]] = None, memory_space: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return AllocTensorOp(result=result, dynamic_sizes=dynamic_sizes, copy=copy, size_hint=size_hint, memory_space=memory_space, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CloneOp(_ods_ir.OpView):
  r"""
  Clones the data in the input view into an implicitly defined output view.
  
  Usage:
  
  ```mlir
  %arg1 = bufferization.clone %arg0 : memref<?xf32> to memref<?xf32>
  ```
  
  Valid implementations of this operation may alias the input and output
  views or create an actual copy. Mutating the source or result
  of the clone operation after the clone operation thus leads to undefined
  behavior.
  """

  OPERATION_NAME = "bufferization.clone"

  _ODS_REGIONS = (0, True)

  def __init__(self, output: _ods_ir.Type, input: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(input)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(output)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def input(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CloneOp)
class CloneOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "bufferization.clone"

  @builtins.property
  def input(self) -> _ods_ir.Value:
    return self.operands[0]

def clone(output: _ods_ir.Type, input: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CloneOp(output=output, input=input, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class DeallocOp(_ods_ir.OpView):
  r"""
  This operation deallocates each of the given memrefs if there is no alias
  to that memref in the list of retained memrefs and the corresponding
  condition value is set. This condition can be used to indicate and pass on
  ownership of memref values (or in other words, the responsibility of
  deallocating that memref). If two memrefs alias each other, only one will be
  deallocated to avoid double free situations.
  
  The number of variadic `memref` operands (the memrefs to be deallocated)
  must equal the number of variadic `condition` operands and correspond to
  each other element-wise.
  
  The `memref` operands must be the originally allocated memrefs, however, the
  `retained` memref operands may be arbitrary memrefs.
  
  This operation returns a variadic number of `updatedConditions` operands,
  one updated condition per retained memref. An updated condition indicates
  the ownership of the respective retained memref. It is computed as the
  disjunction of all `conditions` operands where the corresponding to
  `memrefs` operand aliases with the retained memref. If the retained memref
  has no aliases among `memrefs`, the resulting updated condition is 'false'.
  This is because all memrefs that need to be deallocated within one basic
  block should be added to the same `bufferization.dealloc` operation at the
  end of the block; if no aliasing memref is present, then it does not have to
  be deallocated and thus we don't need to claim ownership. If the memrefs to
  be deallocated are split over multiple dealloc operations (e.g., to avoid
  aliasing checks at runtime between the `memref` operands), then the results
  have to be manually combined using an `arith.ori` operation and all of them
  still require the same list of `retained` memref operands unless the
  (potentially empty) set of aliasing memrefs can be determined statically. In
  that case, the `updatedCondition` operand can be replaced accordingly (e.g.,
  by a canonicalizer).
  
  Example:
  ```mlir
  %0:3 = bufferization.dealloc (%a0, %a1 : memref<2xf32>, memref<4xi32>)
    if (%cond0, %cond1) retain (%r0, %r1, %r2 : memref<?xf32>, memref<f64>,
    memref<2xi32>)
  ```
  Deallocation will be called on `%a0` if `%cond0` is 'true' and neither
  `%r0`, `%r1`, or `%r2` are aliases of `%a0`. `%a1` will be deallocated when
  `%cond1` is set to 'true' and none of `%r0`, `%r1`, `%r2`, and `%a0` are
  aliases.
  
  Note that this can be an expensive operation if there are many operands that
  cannot be optimized away. The runtime cost of this operation (assuming that
  nothing is optimized away) is `O(|memrefs|^2+|memrefs|*|retained|)`. The
  cost in terms of memory space is `O(|memrefs|+|retained|)`. As a result, it
  is recommended to place it carefully in the IR such that most operands can
  be optimized away by running the `buffer-deallocation-simplification` pass.
  """

  OPERATION_NAME = "bufferization.dealloc"

  _ODS_OPERAND_SEGMENTS = [-1,-1,-1,]

  _ODS_REGIONS = (0, True)

  def __init__(self, memrefs: _Sequence[_ods_ir.Value], conditions: _Sequence[_ods_ir.Value[_ods_ir.IntegerType]], retained: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(_get_op_results_or_values(memrefs))
    operands.append(_get_op_results_or_values(conditions))
    operands.append(_get_op_results_or_values(retained))
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def memrefs(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def conditions(self) -> _ods_ir.OpOperandList[_ods_ir.IntegerType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def retained(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 2)
    return operand_range

  @builtins.property
  def updatedConditions(self) -> _ods_ir.OpResultList[_ods_ir.IntegerType]:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(DeallocOp)
class DeallocOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "bufferization.dealloc"

  @builtins.property
  def memrefs(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def conditions(self) -> _ods_ir.OpOperandList[_ods_ir.IntegerType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def retained(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 2)
    return operand_range

def dealloc(memrefs: _Sequence[_ods_ir.Value], conditions: _Sequence[_ods_ir.Value[_ods_ir.IntegerType]], retained: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, DeallocOp]:
  op = DeallocOp(memrefs=memrefs, conditions=conditions, retained=retained, results=results, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class DeallocTensorOp(_ods_ir.OpView):
  r"""
  `bufferization.dealloc_tensor` is a buffer deallocation in tensor land. This
  op can be used for manual buffer deallocation. Some bufferizations (such as
  One-Shot Bufferize) take care of buffer deallocation, in which case this op
  is usually not needed. Details can be found in the documentation of the
  respective bufferization passes.
  
  In case of a dense tensor, this op lowers to a `memref.dealloc` op during
  bufferization.
  
  In case of a sparse tensor, this op releases the underlying sparse storage
  format for a tensor that materialized earlier through a `new` operation, a
  `convert` operation with annotated destination tensor type (unless the
  convert is folded away), or a `bufferization.alloc_tensor` operation. The
  release operation should only be called once for any materialized tensor.
  After this operation, any subsequent `memref` querying operation on the
  tensor returns undefined results.
  
  Example:
  
  ```mlir
  bufferization.dealloc_tensor %tensor : tensor<1024x1024xf64, #CSR>
  ```
  """

  OPERATION_NAME = "bufferization.dealloc_tensor"

  _ODS_REGIONS = (0, True)

  def __init__(self, tensor: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(tensor)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def tensor(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(DeallocTensorOp)
class DeallocTensorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "bufferization.dealloc_tensor"

  @builtins.property
  def tensor(self) -> _ods_ir.Value:
    return self.operands[0]

def dealloc_tensor(tensor: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> DeallocTensorOp:
  return DeallocTensorOp(tensor=tensor, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class MaterializeInDestinationOp(_ods_ir.OpView):
  r"""
  This op indicates that the data of the `source` tensor is guaranteed to
  materialize in `dest`, which can be a tensor or a memref. In case of a
  tensor, `source` materializes in the future buffer of `dest` and a the
  updated destination tensor is returned. If this is not possible, e.g.,
  because the destination tensor is read-only or because its original
  contents are still read later, the input IR fails to bufferize. In case of a
  memref, `source` materializes in `dest`, which is already a buffer. The op
  has no results in that case.
  
  `source`, `dest` and `result` (if present) must have the same runtime shape
  and element type. If the op has a result, the types of `result` and `dest`
  must match exactly (e.g., including any tensor encodings).
  
  By default, this op bufferizes to a memcpy from the future buffer of the
  `source` tensor to the future buffer of the `dest` tensor or to the `dest`
  buffer. However, transformations such as "empty tensor elimination" may
  rewrite IR such that a computation is performed directly in `dest` and no
  memcpy is needed.
  
  If `dest` is a buffer, the `writable` attribute must be specified and the
  `restrict` keyword can be specified. These attributes have the same meaning
  as the respective attributes of `bufferization.to_tensor`.
  
  `writable` indicates that the `dest` buffer is considered writable. It does
  not make sense to materialize a computation in a read-only buffer, so
  `writable` is required.
  
  `restrict` indicates that there is no `bufferization.to_tensor` op and no
  other `bufferization.materialize_in_destination` op with `dest` (or an alias
  thereof) and "restrict". Only ops with this attribute are considered for
  "empty tensor elimination". As part of empty tensor elimination, a new
  `to_tensor` op with `dest` may be inserted and the `restrict` attribute is
  transferred from this op to the new `to_tensor` op. Having "restrict" on
  this op guarantees that performing empty tensor elimination would not create
  invalid IR (i.e., having multiple `to_tensor restrict` with aliasing
  buffers).
  
  Note: `writable` could be removed from this op because it must always be set
  for memref destinations. This op has that attribute to make clear the
  requirements on the `dest` operand in the op assembly format.
  
  Note: If `dest` is a tensor, `tensor.insert_slice` could be used for the
  same purpose, but since tensor dialect ops only indicate *what* should be
  computed but not *where*, it could fold away, causing the computation to
  materialize in a different buffer.
  """

  OPERATION_NAME = "bufferization.materialize_in_destination"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _Optional[_ods_ir.Type], source: _ods_ir.Value, dest: _ods_ir.Value, *, restrict: _Optional[bool] = None, writable: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(source)
    operands.append(dest)
    _ods_context = _ods_get_default_loc_context(loc)
    if bool(restrict): attributes["restrict"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    if bool(writable): attributes["writable"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    if result is not None: results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def source(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def dest(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def restrict(self) -> bool:
    return "restrict" in self.operation.attributes

  @restrict.setter
  def restrict(self, value):
    if bool(value):
      self.operation.attributes["restrict"] = _ods_ir.UnitAttr.get()
    elif "restrict" in self.operation.attributes:
      del self.operation.attributes["restrict"]

  @restrict.deleter
  def restrict(self):
    del self.operation.attributes["restrict"]

  @builtins.property
  def writable(self) -> bool:
    return "writable" in self.operation.attributes

  @writable.setter
  def writable(self, value):
    if bool(value):
      self.operation.attributes["writable"] = _ods_ir.UnitAttr.get()
    elif "writable" in self.operation.attributes:
      del self.operation.attributes["writable"]

  @writable.deleter
  def writable(self):
    del self.operation.attributes["writable"]

  @builtins.property
  def result(self) -> _Optional[_ods_ir.OpResult]:
    return None if len(self.operation.results) < 1 else self.operation.results[0]

@_ods_cext.register_op_adaptor(MaterializeInDestinationOp)
class MaterializeInDestinationOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "bufferization.materialize_in_destination"

  @builtins.property
  def source(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def dest(self) -> _ods_ir.Value:
    return self.operands[1]

  @builtins.property
  def restrict(self) -> bool:
    return "restrict" in self.attributes

  @builtins.property
  def writable(self) -> bool:
    return "writable" in self.attributes

def materialize_in_destination(result: _Optional[_ods_ir.Type], source: _ods_ir.Value, dest: _ods_ir.Value, *, restrict: _Optional[bool] = None, writable: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, MaterializeInDestinationOp]:
  op = MaterializeInDestinationOp(result=result, source=source, dest=dest, restrict=restrict, writable=writable, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class ToBufferOp(_ods_ir.OpView):
  r"""
  An operation that returns the future buffer of a `tensor`.
  
  ```mlir
  // Result type is memref<4x?xf32, #layout, 0>
  %m = bufferization.to_buffer %t : tensor<4x?xf32> to memref<4x?xf32, #layout, 0>
  ```
  
  This operation is a specialized variant of the built-in
  `unrealized_conversion_cast` and is used to make sure that the IR stays
  valid at any point during the bufferization.
  
  The `read_only` attribute can optionally be set, indicating to the
  bufferization that the buffer returned by this op (or an alias created from
  the returned buffer) will not be written to.
  """

  OPERATION_NAME = "bufferization.to_buffer"

  _ODS_REGIONS = (0, True)

  def __init__(self, buffer: _ods_ir.Type, tensor: _ods_ir.Value, *, read_only: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(tensor)
    _ods_context = _ods_get_default_loc_context(loc)
    if bool(read_only): attributes["read_only"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.append(buffer)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def tensor(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def read_only(self) -> bool:
    return "read_only" in self.operation.attributes

  @read_only.setter
  def read_only(self, value):
    if bool(value):
      self.operation.attributes["read_only"] = _ods_ir.UnitAttr.get()
    elif "read_only" in self.operation.attributes:
      del self.operation.attributes["read_only"]

  @read_only.deleter
  def read_only(self):
    del self.operation.attributes["read_only"]

  @builtins.property
  def buffer(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(ToBufferOp)
class ToBufferOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "bufferization.to_buffer"

  @builtins.property
  def tensor(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def read_only(self) -> bool:
    return "read_only" in self.attributes

def to_buffer(buffer: _ods_ir.Type, tensor: _ods_ir.Value, *, read_only: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return ToBufferOp(buffer=buffer, tensor=tensor, read_only=read_only, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class ToTensorOp(_ods_ir.OpView):
  r"""
  An operation that creates a tensor from a buffer. The result value is a
  tensor-like type that must match the corresponding buffer-like operand as
  per TensorLikeType::verifyCompatibleBufferType(). For builtins (TensorType
  and BaseMemRefType), this means that shapes and element types match between
  the tensor and the buffer.
  
  The opposite of this op is `to_buffer`. Together, these two ops are
  useful for source/target materializations when doing type conversions
  involving tensors and buffers.
  
  Example:
  
  ```mlir
  // Produces a value of tensor<4x?xf32> type.
  %t = bufferization.to_tensor %m : memref<4x?xf32, #layout, 0> to tensor<4x?xf32>
  ```
  
  If the `writable` unit attribute is set, the produced tensor is considered
  "writable" during bufferization. Otherwise, every OpOperand that bufferizes
  to a write to the future buffer of the resulting tensor (or an alias
  thereof) will bufferize out-of-place to prevent emitting any writes to
  `memref` during bufferization.
  
  The `restrict` unit attribute (similar to the C `restrict` keyword)
  indicates that the produced tensor result is the only way for the tensor
  IR to gain access to the `memref` operand (or an alias thereof). E.g.,
  there must be no other `to_tensor` op with the same or with an aliasing
  `memref` operand.
  
  Note: Only `to_tensor` ops with the `restrict` unit attribute are supported
  by One-Shot Bufferize. Other IR is rejected. (To support `to_tensor`
  without `restrict`, One-Shot Bufferize would have to analyze memref IR.)
  Ops that have incorrect usage of `restrict` may bufferize incorrectly.
  
  Example:
  
  ```
  %t = bufferization.to_tensor %m restrict writable : memref<4xf32> to tensor<4xf32>
  
  // %t is writable, so the tensor.insert may bufferize in-place in the
  // absence of other conflicts.
  %r = tensor.insert %f into %t[%idx] : tensor<4xf32>
  ```
  
  `to_tensor` ops are not bufferized. They are expected to fold away after
  bufferization. If there are non-bufferizable ops in the IR and
  `allowUnknownOps` is set, they may be part of the resulting IR and not fold
  away. However, such IR is no longer bufferizable with One-Shot Bufferize.
  """

  OPERATION_NAME = "bufferization.to_tensor"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, buffer: _ods_ir.Value, *, restrict: _Optional[bool] = None, writable: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(buffer)
    _ods_context = _ods_get_default_loc_context(loc)
    if bool(restrict): attributes["restrict"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    if bool(writable): attributes["writable"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def buffer(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def restrict(self) -> bool:
    return "restrict" in self.operation.attributes

  @restrict.setter
  def restrict(self, value):
    if bool(value):
      self.operation.attributes["restrict"] = _ods_ir.UnitAttr.get()
    elif "restrict" in self.operation.attributes:
      del self.operation.attributes["restrict"]

  @restrict.deleter
  def restrict(self):
    del self.operation.attributes["restrict"]

  @builtins.property
  def writable(self) -> bool:
    return "writable" in self.operation.attributes

  @writable.setter
  def writable(self, value):
    if bool(value):
      self.operation.attributes["writable"] = _ods_ir.UnitAttr.get()
    elif "writable" in self.operation.attributes:
      del self.operation.attributes["writable"]

  @writable.deleter
  def writable(self):
    del self.operation.attributes["writable"]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(ToTensorOp)
class ToTensorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "bufferization.to_tensor"

  @builtins.property
  def buffer(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def restrict(self) -> bool:
    return "restrict" in self.attributes

  @builtins.property
  def writable(self) -> bool:
    return "writable" in self.attributes

def to_tensor(result: _ods_ir.Type, buffer: _ods_ir.Value, *, restrict: _Optional[bool] = None, writable: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return ToTensorOp(result=result, buffer=buffer, restrict=restrict, writable=writable, loc=loc, ip=ip).result



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from .._gpu_transform_ops_gen import *
from .._gpu_transform_ops_gen import _Dialect

try:
    from ...ir import *
    from ...dialects import transform
    from .._ods_common import _cext as _ods_cext
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from typing import Optional, Sequence, Union, overload


@_ods_cext.register_operation(_Dialect, replace=True)
class MapForallToBlocks(MapForallToBlocks):
    """Specialization for MapForallToBlocks class."""

    @overload
    def __init__(
        self,
        result_type: Type,
        target: Union[Operation, OpView, Value],
        *,
        grid_dims: Optional[Union[Sequence[int], Attribute]] = None,
        generate_gpu_launch: Optional[Union[bool, Attribute]] = None,
        loc=None,
        ip=None,
    ):
        ...

    @overload
    def __init__(
        self,
        target: Union[Operation, OpView, Value],
        *,
        grid_dims: Optional[Union[Sequence[int], Attribute]] = None,
        generate_gpu_launch: Optional[Union[bool, Attribute]] = None,
        loc=None,
        ip=None,
    ):
        ...

    def __init__(
        self,
        result_type_or_target: Union[Operation, OpView, Type, Value],
        target_or_none: Optional[Union[Operation, OpView, Value]] = None,
        *,
        grid_dims: Optional[Union[Sequence[int], Attribute]] = None,
        generate_gpu_launch: Optional[Union[bool, Attribute]] = None,
        loc=None,
        ip=None,
    ):
        if isinstance(result_type_or_target, Type):
            result_type = result_type_or_target
            target = target_or_none
        else:
            result_type = transform.AnyOpType.get()
            target = result_type_or_target

        super().__init__(
            result_type,
            target,
            grid_dims=grid_dims,
            generate_gpu_launch=generate_gpu_launch,
            loc=loc,
            ip=ip,
        )


@_ods_cext.register_operation(_Dialect, replace=True)
class MapNestedForallToThreads(MapNestedForallToThreads):
    """Specialization for MapNestedForallToThreads class."""

    @overload
    def __init__(
        self,
        result_type: Type,
        target: Union[Operation, OpView, Value],
        *,
        block_dims: Optional[Sequence[int]] = None,
        warp_size: Optional[Sequence[int]] = None,
        sync_after_distribute: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        ...

    @overload
    def __init__(
        self,
        target: Union[Operation, OpView, Value],
        *,
        block_dims: Optional[Sequence[int]] = None,
        warp_size: Optional[Sequence[int]] = None,
        sync_after_distribute: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        ...

    def __init__(
        self,
        result_type_or_target: Union[Operation, OpView, Value, Type],
        target_or_none: Optional[Union[Operation, OpView, Value]] = None,
        *,
        block_dims: Optional[Union[Sequence[int], Attribute]] = None,
        warp_size: Optional[Union[Sequence[int], Attribute]] = None,
        sync_after_distribute: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        if isinstance(result_type_or_target, Type):
            result_type = result_type_or_target
            target = target_or_none
        else:
            result_type = result_type_or_target.type
            target = result_type_or_target
        super().__init__(
            result_type,
            target,
            block_dims=block_dims,
            warp_size=warp_size,
            sync_after_distribute=sync_after_distribute,
            loc=loc,
            ip=ip,
        )



# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
from contextlib import contextmanager
from typing import Optional, Sequence, Union

from .arith import constant as _ext_arith_constant, index_cast
from numba_cuda_mlir._mlir.extras.meta import region_op
from ..util import region_adder
from numba_cuda_mlir._mlir.dialects._ods_common import (
    _cext,
    get_default_loc_context,
    get_op_result_or_op_results,
)

# gotta come first
from numba_cuda_mlir._mlir.dialects.scf import *
from numba_cuda_mlir._mlir.dialects.scf import _Dialect, yield_ as yield__
from numba_cuda_mlir._mlir.ir import (
    Attribute,
    IndexType,
    InsertionPoint,
    OpResultList,
    OpView,
    OpaqueType,
    Operation,
    Value,
    _denseI64ArrayAttr,
)

opaque = lambda dialect_namespace, buffer: OpaqueType.get(dialect_namespace, buffer)


def canonicalize_start_stop_step(start, stop, step):
    if step is None:
        step = 1
    if stop is None:
        stop = start
        start = 0
    params = [start, stop, step]
    type = IndexType.get()
    maybe_types = {p.type for p in params if isinstance(p, Value)}
    if maybe_types:
        if len(maybe_types) > 1:
            raise ValueError(
                f"all {start=} and {stop=} and {step=} ir.Value objects must have the same type"
            )
        type = maybe_types.pop()

    for i, p in enumerate(params):
        if isinstance(p, int):
            p = _ext_arith_constant(p, type=type)
        assert isinstance(p, Value)
        params[i] = p

    return params[0], params[1], params[2]


def _build_for(
    start,
    stop=None,
    step=None,
    iter_args: Optional[Sequence[Value]] = None,
    *,
    loc=None,
    ip=None,
):
    start, stop, step = canonicalize_start_stop_step(start, stop, step)
    return ForOp(start, stop, step, iter_args, loc=loc, ip=ip)


def range_(
    start,
    stop=None,
    step=None,
    iter_args: Optional[Sequence[Value]] = None,
    *,
    loc=None,
    ip=None,
):
    for_op = _build_for(start, stop, step, iter_args, loc=loc, ip=ip)
    iv = for_op.induction_variable
    iter_args = tuple(for_op.inner_iter_args)
    with InsertionPoint(for_op.body):
        if len(iter_args) > 1:
            yield iv, iter_args, for_op.results
        elif len(iter_args) == 1:
            yield iv, iter_args[0], for_op.results[0]
        else:
            yield iv


def placeholder_opaque_t():
    return opaque("scf", "placeholder")


for__ = region_op(_build_for, terminator=yield__)


def _parfor(op_ctor):
    def _base(lower_bounds, upper_bounds=None, steps=None, *, loc=None, ip=None, **kwargs):
        if upper_bounds is None:
            upper_bounds = lower_bounds
            lower_bounds = [0] * len(upper_bounds)
        if steps is None:
            steps = [1] * len(lower_bounds)

        params = [list(lower_bounds), list(upper_bounds), list(steps)]
        for i, p in enumerate(params):
            for j, pp in enumerate(p):
                if isinstance(pp, int):
                    pp = _ext_arith_constant(pp, index=True)
                if not isinstance(pp.type, IndexType):
                    pp = index_cast(pp)
                p[j] = pp
            params[i] = p

        return op_ctor(*params, loc=loc, ip=ip, **kwargs)

    return _base


@region_op
def in_parallel():
    return InParallelOp()


def in_parallel_(parallel_insert_slice=None):
    if isinstance(parallel_insert_slice, (tuple, list)):
        assert len(parallel_insert_slice) <= 1, "expected at most one parallel_insert_slice op"
        if len(parallel_insert_slice) == 1:
            parallel_insert_slice = parallel_insert_slice[0]
        else:
            parallel_insert_slice = None

    @in_parallel
    def foo():
        if parallel_insert_slice is not None:
            parallel_insert_slice()
        return


def parallel_insert_slice(
    source,
    dest,
    static_offsets=None,
    static_sizes=None,
    static_strides=None,
    offsets=None,
    sizes=None,
    strides=None,
):
    from . import tensor

    @in_parallel
    def foo():
        tensor.parallel_insert_slice(
            source,
            dest,
            offsets,
            sizes,
            strides,
            static_offsets,
            static_sizes,
            static_strides,
        )


forall_ = region_op(_parfor(ForallOp), terminator=in_parallel_)


def _parfor_context_manager(op_ctor):
    def _base(*args, **kwargs):
        for_op = _parfor(op_ctor)(*args, **kwargs)
        block = for_op.regions[0].blocks[0]
        block_args = tuple(block.arguments)
        with InsertionPoint(block):
            yield block_args

    return _base


forall = _parfor_context_manager(ForallOp)


class ParallelOp(ParallelOp):
    def __init__(
        self,
        lower_bounds,
        upper_bounds,
        steps,
        *,
        inits: Optional[Union[Operation, OpView, Sequence[Value]]] = None,
        loc=None,
        ip=None,
    ):
        assert len(lower_bounds) == len(upper_bounds) == len(steps)
        if inits is None:
            inits = []
        results = [i.type for i in inits]
        iv_types = [IndexType.get()] * len(lower_bounds)
        super().__init__(
            results,
            lower_bounds,
            upper_bounds,
            steps,
            inits,
            loc=loc,
            ip=ip,
        )
        self.regions[0].blocks.append(*iv_types)

    @property
    def body(self):
        return self.regions[0].blocks[0]

    @property
    def induction_variables(self):
        return self.body.arguments


def _parallel_terminator(xs):
    if len(xs):
        raise ValueError(
            "default return->parallel terminator does not support operands; use scf.reduce_ instead"
        )
    return reduce_()


parallel_ = region_op(_parfor(ParallelOp), terminator=_parallel_terminator)
parallel = _parfor_context_manager(ParallelOp)


def while___(cond: Value, *, loc=None, ip=None):
    def wrapper():
        nonlocal ip
        inits = list(cond.owner.operands)
        results_ = [i.type for i in inits]
        while_op = WhileOp(results_, inits, loc=loc, ip=ip)
        while_op.regions[0].blocks.append(*[i.type for i in inits])
        before = while_op.regions[0].blocks[0]
        while_op.regions[1].blocks.append(*[i.type for i in inits])
        after = while_op.regions[1].blocks[0]
        with InsertionPoint(before) as ip:
            cond_ = condition(cond, list(before.arguments))
            cond.owner.move_before(cond_)
        with InsertionPoint(after):
            yield inits

    if hasattr(while___, "wrapper"):
        # needed in order to exit the `after` insertion point
        next(while___.wrapper, False)
        del while___.wrapper
        return False
    else:
        while___.wrapper = wrapper()
        # enter `after` insertion point
        return next(while___.wrapper)


def while__(cond: Value, *, loc=None, ip=None):
    yield while___(cond, loc=loc, ip=ip)
    yield while___(cond, loc=loc, ip=ip)


class ReduceOp(ReduceOp):
    def __init__(self, operands, num_reductions, *, loc=None, ip=None):
        super().__init__(operands, num_reductions, loc=loc, ip=ip)
        for i in range(num_reductions):
            self.regions[i].blocks.append(operands[i].type, operands[i].type)


def reduce_(*operands, loc=None, ip=None):
    return ReduceOp(operands, len(operands), loc=loc, ip=ip)


reduce = region_op(reduce_, terminator=lambda xs: reduce_return(*xs))


@region_adder(terminator=lambda xs: reduce_return(*xs))
def another_reduce(reduce_op):
    for r in reduce_op.regions:
        if len(r.blocks[0].operations) == 0:
            return r


def yield_(*args, results_=None):
    if len(args):
        assert results_ is None, "must provide results_ or args"
    if results_ is not None:
        args = results_
    if len(args) == 1 and isinstance(args[0], (list, OpResultList)):
        args = list(args[0])
    y = yield__(args)
    parent_op = y.operation.parent.opview
    if len(parent_op.results):
        results = get_op_result_or_op_results(parent_op)
        assert (
            isinstance(results, (OpResultList, Value))
            or isinstance(results, list)
            and all(isinstance(r, Value) for r in results)
        ), f"api has changed: {results=}"
        if isinstance(results, Value):
            results = [results]
        unpacked_args = args
        if any(isinstance(a, OpResultList) for a in unpacked_args):
            assert len(unpacked_args) == 1
            unpacked_args = list(unpacked_args[0])

        for i, r in enumerate(results):
            if r.type == placeholder_opaque_t():
                r.set_type(unpacked_args[i].type)

        if len(results) > 1:
            return results
        return results[0]
    elif len(args):
        raise RuntimeError(f"can't yield from parent_op which has no results")
    return None


def _if(cond, results=None, *, has_else=False, loc=None, ip=None):
    if results is None:
        results = []
    if results:
        has_else = True
    return IfOp(cond, results, has_else=has_else, loc=loc, ip=ip)


if_ = region_op(_if, terminator=yield__)


@contextmanager
def if_ctx_manager(cond, results=None, *, has_else=False, loc=None, ip=None):
    if_op = _if(cond, results, has_else=has_else, loc=loc, ip=ip)
    with InsertionPoint(if_op.regions[0].blocks[0]):
        yield if_op


@contextmanager
def else_ctx_manager(if_op):
    if len(if_op.regions[1].blocks) == 0:
        if_op.regions[1].blocks.append(*[])
    with InsertionPoint(if_op.regions[1].blocks[0]):
        yield


@region_adder(terminator=yield__)
def else_(ifop):
    return ifop.regions[1]


execute_region = region_op(ExecuteRegionOp)
